"""
CS 1
Version: 11/04/23 23fa
MP6 Testing Program to test Part B/C with import/export functionality.
Make sure to use this test _only_ after testing test_isolated_mp6_ab.py.
See instructions.txt for more details.
"""
# Import student classes/functions
from lattice import *
from lattice_client import *
from atom import Atom

# Imports for testing
import os
import random
import pytest

TEST_COORDS_1 = [(0.0, 0.0, 0.0), (1.0, 1.0, 1.0), (.5, .5, .5)]

# Test cases for atoms
ATOM_ORIGIN = [Atom('Ir', (0.0, 0.0, 0.0))]
ATOM_CORNER = [Atom('Au', (1.0, 1.0, 1.0))]
ATOM_TWO_AXES = [Atom('Pt', (0.0, 1.0, 1.0))]
ATOM_ONE_AXIS = [Atom('Y', (1.0, 0.0, 0.0))]
ATOM_CENTER = [Atom('Fe', (.5, .5, .5))]
ATOM_CORNER_DOUBLE_ATOM_FACE = [
    Atom('H', (1.0, 1.0, 1.0)),
    Atom('O', (1.0, 1.0, 0.0)),
    Atom('O', (1.0, 0.0, 1.0))
]

# Generalize a list of different atom structures to test
ATOM_INPUTS = [
    ATOM_ORIGIN, ATOM_CORNER, ATOM_TWO_AXES, 
    ATOM_CORNER_DOUBLE_ATOM_FACE,
    ATOM_CENTER
]


# Testing helper functions
def _gen_lattice_param_test_inp(a_list):
    """
    Generates parameters for a random test lattice given an atom list `a_list`. 

    Returns:
      - ([Atom], int) - (atoms, lparam) parameters for lattice to construct.
    """
    return (a_list, random.randint(1.0, 50))


# This notation is a fancy pytest feature to parameterize tests; students
# are not required to know this notation.
@pytest.mark.parametrize('atom_list, lp', [_gen_lattice_param_test_inp(inp)
                                           for inp in ATOM_INPUTS])
def test_get_lattice_parameter(atom_list, lp):
    # Tests for lattice parameters being initialized for each new
    # CubicLattice (requires __init__ and get_lattice_parameter to pass)
    assert CubicLattice(lp, atom_list).get_lattice_parameter() == lp


@pytest.mark.parametrize('atom_list,lp', [_gen_lattice_param_test_inp(inp)
                                           for inp in ATOM_INPUTS])
def test_cell_volume(atom_list, lp):
    # Tests for B.3. get_cell_volume method
    # CubicLattice (requires __init__ to pass)
    assert CubicLattice(lp, atom_list).get_cell_volume() == lp * lp * lp


@pytest.mark.parametrize('atom_list, expected_elements', [(ATOM_ORIGIN, ['Ir']), (ATOM_CORNER_DOUBLE_ATOM_FACE, ['H', 'O']), ([*ATOM_ORIGIN, *ATOM_CORNER, *ATOM_ORIGIN, *ATOM_CORNER], ['Ir', 'Au'])])
def test_get_unique_elements(atom_list, expected_elements): 
    # Tests for B.4. get_unique_elements method
    # CubicLattice (requires __init__ and get_lattice_parameter to pass)
    assert CubicLattice(1.0, atom_list).get_unique_elements() == expected_elements


@pytest.mark.parametrize('atom_list, expected_coord', [(ATOM_CORNER, ATOM_ORIGIN), (ATOM_TWO_AXES, ATOM_ONE_AXIS)])
def test_inverted_cell(atom_list, expected_coord):
    # Tests for B.6. get_inverted_cell method
    # CubicLattice (requires previous methods to be correctly implemented)
    og_crystal = CubicLattice(1.0, atom_list)
    original_atoms = og_crystal.atoms
    inverted_crystal = og_crystal.get_inverted_cell()
    assert inverted_crystal.atoms[0].coordinates == expected_coord[0].coordinates
    # The inverted CubicLattice result should have the same number of atoms
    # as the original
    assert og_crystal.atoms == original_atoms # Assure no overwrite


@pytest.mark.parametrize('atom_list, lp, expected',
                         [(ATOM_CORNER, 2.0, 1.0), (ATOM_TWO_AXES, 2.0, .5),
                          (ATOM_CORNER_DOUBLE_ATOM_FACE,  2.0, 2.0)])
def test_number_atoms(atom_list, lp, expected):
    # Tests for B.5. get_number_atoms method
    # CubicLattice (requires previous methods to be correctly implemented)
    assert CubicLattice(lp, atom_list).get_number_atoms() == expected 


# Test Suite 2: File Import/Export Functionality
# Program constants for given xyz files for example crystals
IRON_BCC_FNAME = 'iron_bcc_out.xyz'
IRON_BCC = create_BCC('Fe', 3.5) 

ALUMINUM_FCC_FNAME = 'aluminum_fcc_out.xyz'
ALUMINUM_FCC = create_FCC('Al', 4) 

POLONIUM_SC_FNAME = 'polonium_sc_out.xyz'
POLONIUM_SC = create_SC('Po', 3.34)

@pytest.mark.parametrize('crystal,crystal_name,act_filename', [(IRON_BCC, 'Iron BCC', IRON_BCC_FNAME), (ALUMINUM_FCC, 'Aluminum FCC', ALUMINUM_FCC_FNAME), (POLONIUM_SC, 'Polonium SC', POLONIUM_SC_FNAME)])
def test_create_and_write(crystal, crystal_name, act_filename):
    # Tests for Part C file-writing (exporting)
    # Note to students: If you see any "file-not-found" related errors for this test,
    # make sure to double-check instructions.txt
    crystal.write_to_xyz(crystal_name, os.path.join('student_files/', act_filename))
    
    with open(os.path.join('test_files/', f'{act_filename}')) as test_dat: 
        # Note: readlines() is only appropriate for these tests and shouldn't 
        # be used elsewhere
        test_out = test_dat.readlines() 

    standardized_stud_out = ''
    with open(os.path.join('student_files', f'{act_filename}')) as stud_dat: 
        stud_out = stud_dat.readlines()
        # First, check that the number of lines in the files are equal
        assert len(stud_out) == len(test_out)

        # Check that # of atoms are the same (the first line in the file)
        assert test_out[0] == stud_out[0]

        # Check that crystal name is the same (the second line in the file)
        assert test_out[1] == stud_out[1]

        # The following standardizes both outs to floats 
        # (accepts 0 vs. 0.0, but requires 2.5 not 2.0)

        # Remove first two lines already tested
        test_out = test_out[2:]
        stud_out = stud_out[2:]

        standardized_stud_out = []
        standardized_test_out = []

        # Here, we standardize lines to use floats for all coordinate values
        for i, line in enumerate(stud_out):
            expected_line = test_out[i] 
            # Each atom line should have exactly 4 components 'el x y z'
            assert line.count(' ') == 3

            # Standardize student output comparison to floats
            # Get actual values for each atom line (tuple-unpacking!)
            (element_act, x_act, y_act, z_act) = line.split()

            # Get expected values for each atom line
            (element_exp, x_exp, y_exp, z_exp) = expected_line.split()

            # Add both results to the standardized lists
            standardized_stud_out.append(f'{element_act} {float(x_act)} {float(y_act)} {float(z_act)}\n')
            standardized_test_out.append(f'{element_exp} {float(x_exp)} {float(y_exp)} {float(z_exp)}\n')

        # Now that everything's standardized, sorting can be used to
        # assert expected xyz representation
        standardized_stud_out.sort()
        standardized_test_out.sort()

        assert standardized_stud_out == standardized_test_out


@pytest.mark.parametrize('in_filename, reference', [(IRON_BCC_FNAME, IRON_BCC), (ALUMINUM_FCC_FNAME, ALUMINUM_FCC), (POLONIUM_SC_FNAME, POLONIUM_SC)])
def test_read_xyz(in_filename, reference):
    # Tests for Part C read_xyz (importing)
    # Requires both the read_xyz method _and_ the optional in_filename __init__
    # parameter to be correctly implemented
    created_crystal = CubicLattice(1.0, in_filename=f'test_files/{in_filename}')
    # The following use "list comprehension" with is a topic we haven't seen yet
    # and shouldn't be used without permission (since students usually use these
    # inefficiently). Build a list of coordinates for each atom in the crystals
    # and convert to float values for improved testing
    actual_coords = [atom.coordinates for atom in created_crystal.atoms] 
    actual_coords = [(float(x), float(y), float(z)) for (x, y, z) in actual_coords] 
    exp_coords = [atom.coordinates for atom in reference.atoms]
    exp_coords = [(float(x), float(y), float(z)) for (x, y, z) in exp_coords] 

    actual_coords.sort()
    exp_coords.sort()
    assert actual_coords == exp_coords
